/*
 * File: process.c
 * Date: 2011-12-14
 *
 * author: Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 * MACOS port by Aria Stewart <aredridel@nbtsc.org>
 */


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<sys/stat.h>
#include	"rho.h"
#include	"buffer.h"
#include	"cursor.h"
#include	"file.h"
#include	"fileio.h"
#include	"line.h"
#include	"macros.h"
#include	"messages.h"
#include	"process.h"
#include	"terminal.h"
#include	"utils.h"
#include	"window.h"


static const char seps[] = " :";        /* used by make and *_error */


/*++

Spawn a command. execute a program

++*/
int shell_command(f, n)
int f, n;
{
    char line[NLINE];
    int s;

    f = n = 0;                  /* splint */

    /* Get command line */
    s = mlreply("!", line, NLINE);
    if (!s)
        return (s);

    /* Perform the command */
    s = callout(line);

    /* if we are interactive, pause here */
    if (clexec == FALSE) {
        ml_end();
    }
    return (s);
}


/*++

Load results of "grep -l <user input>"

++*/
int file_grep(f, n)
int f, n;
{

    char *fname;
    char grep[NLINE] = "grep -l ";
    char line[NLINE];
    char outfile[] = RHOOUT;
    int s;
    LINE *lp;
    BUFFER *bp;

    f = n = 0;                  /* splint */

    /* Get command line */
    s = mlreply("grep -l ", line, NLINE);
    if (!s)
        return (s);

    strcat(grep, line);
    strcat(grep, ">");
    strcat(grep, outfile);

    /* Perform the command */
    s = callout(grep);
    if (!s || (!getfile(outfile, TRUE))) {
        mlwrite("NO RESULTS");
        ttgetc();
        return FALSE;
    }

    while (s) {
        lp = curwp->w_dotp;
        fname = strdup(ltext(lp));
        fname[lused(lp)] = '\0';
        if (strcmp(fname, "") == 0)
            break;
        if (fname[strlen(fname) - 1] == '~') {
            s = next_line(0, 1);
            continue;
        }
        bp = exist_buffer(fname, TRUE, 0);
        bp->b_active = FALSE;
        s = next_line(0, 1);
    }

    make_bp_top(bp);
    delbufbyname(outfile);
    remove(outfile);
    return (s);
}



/*++

Pipe output of program to buffer

++*/
int pipe_command(f, n)
int f, n;
{
    BUFFER *bp;
    char line[NLINE];
    char outfile[] = RHOOUT;
    char errfile[] = RHOERR;
    int s = 0;
    struct stat st;
    struct stat st2;

    f = n = 0;                  /* splint */

    /* Get pipe-in command */
    s = mlreply("@", line, NLINE);
    if (!s) {
        return (s);
    }

    /* Get rid of old output buffers */
    single_window(0, 0);

    bp = exist_buffer(outfile, FALSE, 0);
    if (bp) {
        bpflip_release(bp);
    }
    bp = exist_buffer(errfile, FALSE, 0);
    if (bp) {
        bpflip_release(bp);
    }

    /* Add redirects and callout */
    strcat(line, ">");
    strcat(line, outfile);
    strcat(line, " 2>");
    strcat(line, errfile);
    callout(line);

    /* load stderr else stdout */
    if (!split_window(TRUE, POPUPS))
        return (0);
    next_window(FALSE, 0);

    (void) stat(errfile, &st);
    if (st.st_size > 0) {
        if (!getfile(errfile, FALSE))
            return (0);
        window_ro();
    }
    (void) stat(outfile, &st2);
    if (st2.st_size > 0) {
        if (!getfile(outfile, FALSE))
            return (0);
        window_ro();
    }

    remove(outfile);
    remove(errfile);
    return (1);
}



/*++

Interactive wrapper for filterbuffer

++*/
int filter_buffer(f, n)
int f, n;
{

    int s;
    char line[NLINE];

    f = n = 0;                  /* splint */

    /* Get the filter name and its args */
    s = mlreply("#", line, NLINE);
    if (!s || (s == ABORT))
        return (s);
    else
        return (filterbuffer((char *) line));
}


/*++

Filter buffer through command

++*/
int filterbuffer(line)
char *line;
{
    int s, t;
    char tmpnam[LENPATH];
    BUFFER *bp;
    static char filnam1[] = "fltinp";
    static char filnam2[] = "fltout";
    struct stat st;

    /* Don't allow filtering of VIEW mode buffer */
    if (curbp->b_mode & MDVIEW)
        return (err_readonly());

    /* Setup the proper file names */
    bp = curbp;
    strcpy(tmpnam, bp->b_name); /* Save the original name */
    strcpy(bp->b_name, filnam1);        /* Set it to our new one */

    /* Write it out, checking for errors */
    if (!writeout(filnam1, "w")) {
        mlwrite("[Cannot write filter file]");
        strcpy(bp->b_name, tmpnam);
        return (0);
    }

    /* Setup input and output */
    strcat(line, " <fltinp >fltout");

    /* Perform command */
    s = callout(line);
    (void) stat(filnam2, &st);

    /* If successful, read in file */
    if ((s) && (st.st_size > 0)) {
        t = readin(filnam2, FALSE);
        /* Mark buffer as changed */
        if (t)
            bp->b_flag |= BFCHG;
    } else
        mlwrite("[Execution failed]");

    /* Reset file name */
    strcpy(bp->b_name, tmpnam);

    /* rm temporary files */
    remove(filnam1);
    remove(filnam2);

    return (s);
}


/*++

Helper for make and next_error

++*/
static void nexterror(current)
char *current;
{
    int status = TRUE;
    LINE *lp;

    lp = curwp->w_dotp;
    strcpy(current, ltext(lp));
    while (strstr(current, "error:") == NULL &&
           strstr(current, "warning:") == NULL) {
        status = next_line(0, 1);
        if (status == FALSE)
            break;
        lp = curwp->w_dotp;
        strcpy(current, ltext(lp));
    }
    status = next_line(0, 1);

}


/*++

Execute compiler, generating _rho_stderr file.  On error, load error file
in popup, problem file in main window with cursor on error/warning line.

++*/
int make()
{

    BUFFER *bp;
    BUFFER *origbp = curbp;
    char *cmd;
    char current[NLINE];
    char *efile;
    char *eline;
    char errfile[] = RHOERR;
    int f = FALSE;
    int n = ERRWIN;
    struct stat st;

    /* build and spawn compiler cmd  */
    cmd = strcat(compiler, " 2> ");
    cmd = strcat(cmd, RHOERR);
    callout(cmd);

    single_window(0, 0);

    /* new errors if stderr */
    (void) stat(RHOERR, &st);
    if (st.st_size > 0) {

        /* get rid of old errors */
        bp = exist_buffer("scratch", TRUE, 0);
        make_bp_top(bp);
        delbufbyname(RHOERR);

        /* stderr ro in popup */
        split_window(f, n);
        readin(errfile, TRUE);
        window_ro();

        /* problem at cursor */
        nexterror(current);
        if (strstr(current, "error") != NULL ||
            strstr(current, "warning") != NULL) {
            efile = strtok(current, seps);
            eline = strtok(NULL, seps);
            previous_window(f, 0);
            getfile(efile, TRUE);
            goto_line(TRUE, str2int(eline));
        } else {                /* restore original buffer */
            previous_window(f, 0);
            make_bp_top(origbp);
        }

    } else
        ml_end();

    remove(RHOERR);
    return (TRUE);
}



/*++

Goto the next problem line indicated in the stderr window.

++*/
int next_error()
{

    BUFFER *origbp = curbp;
    char current[NLINE];
    char *efile;
    char *eline;
    int f = FALSE;
    int n = FALSE;

    /* find next problem */
    previous_window(f, n);
    nexterror(current);
    if (strstr(current, "error") != NULL ||
        strstr(current, "warning") != NULL) {
        efile = strtok(current, seps);
        eline = strtok(NULL, seps);

        /* problem at cursor */
        previous_window(f, n);
        if (strcmp(efile, curbp->b_name) != 0)
            mlputs(TEXT25);     /* Re-run compiler for next file */
        else
            goto_line(TRUE, str2int(eline));
    } else {                    /* restore original buffer */
        previous_window(f, 0);
        make_bp_top(origbp);
    }

    return (TRUE);
}



/*++

Run "make clean"

++*/
int make_clean()
{
    int status = 0;

    status = callout("make clean");
    ml_end();

    return (status);
}



/*++

Goto file and problem line indicated at the cursor.
Assumes line begins "filename:linenumber:" or "filename: line linenumber:"

++*/
int this_error()
{

    char current[NLINE];
    char *efile;
    char *eline;
    int f = FALSE;
    int n = FALSE;
    LINE *lp;
    struct stat st;
    char *script;

    lp = curwp->w_dotp;
    strcpy(current, ltext(lp));
    efile = strtok(current, seps);

    if (!fexist(efile))
        return (0);

    if (strstr(efile, "_rho_") != NULL) {
        script = &efile[strlen(efile)];
        while (*script != '_')
            script--;
        script++;
        efile = script;
    }

    /* problem at cursor */
    (void) stat(efile, &st);
    if (st.st_size > 0) {
        eline = strtok(NULL, seps);     /* compiler style errors */
        if (strcmp(eline, "line") == 0)
            eline = strtok(NULL, seps); /* bash style errors */
        previous_window(f, n);
        if (!getfile(efile, TRUE))
            return (0);
        goto_line(TRUE, str2int(eline));
    } else {
        return (FALSE);
    }
    return (TRUE);
}

/*++

Create new buffer with ~/.rho/plate.kind else $bangline

++*/
int create_script()
{

    BUFFER *bp;
    char kind[8];
    char name[NLINE];
    char plate[LENPATH];
    int s;

    /* Get the script kind */
    s = mlreply(TEXT29, kind, NLINE);
    if (!s || (s == ABORT))
        return (s);

    /* Get the script name */
    s = mlreply(TEXT28, name, NLINE);
    if (!s || (s == ABORT))
        return (s);

    /* add ext if needed */
    if (strstr(name, kind) == NULL) {
        strcat(name, ".");
        strcat(name, kind);
    }

    /* create buffer for script */
    bp = exist_buffer(name, TRUE, 0);

    make_bp_top(bp);

    /* add plate else bangline */
    strcpy(plate, getenv("HOME"));
    strcat(plate, "/.rho/plate.");
    strcat(plate, kind);
    if (fexist(plate)) {
        insertfile(plate);
    } else {
        insertstring(bangline);
        newline(0, 2);
    }

    return (TRUE);
}



/*++

Execute script a buffer created by create_script

++*/
int execute_script()
{

    BUFFER *bp;
    char errfile[] = RHOERR;
    char outfile[] = RHOOUT;
    char script[LENPATH];
    char tmpscr[LENPATH];
    struct stat st2;
    struct stat st;

    /* Get rid of old output buffers */
    single_window(0, 0);

    bp = exist_buffer(outfile, FALSE, 0);
    if (bp) {
        bpflip_release(bp);
    }
    bp = exist_buffer(errfile, FALSE, 0);
    if (bp) {
        bpflip_release(bp);
    }

    /* Write out current buf */
    strcpy(tmpscr, "_rho_");
    strcat(tmpscr, curbp->b_name);
    if (!writeout(tmpscr, "w")) {
        mlwrite("[Cannot write script]");
        return (0);
    } else {
#if MACOS
        chmod(tmpscr, (mode_t) S_IRWXU);
#else
        chmod(tmpscr, (__mode_t) S_IRWXU);
#endif                          /* MACOS */
    }

    /* Add redirects and callout */
    strcpy(script, tmpscr);
    strcat(script, " >");
    strcat(script, outfile);
    strcat(script, " 2>");
    strcat(script, errfile);
    callout(script);

    /* load stderr else stdout */
    if (!split_window(TRUE, POPUPS))
        return (0);
    next_window(FALSE, 0);

    (void) stat(errfile, &st);
    if (st.st_size > 0) {
        if (!getfile(errfile, FALSE))
            return (0);
        window_ro();
    } else {
        (void) stat(outfile, &st2);
        if (st2.st_size > 0) {
            if (!getfile(outfile, FALSE))
                return (0);
            window_ro();
        } else {
            ml_end();           /* should not get here */
        }
    }

    remove(tmpscr);
    remove(outfile);
    remove(errfile);
    return (1);
}
